# Owner(s): ["oncall: distributed"]

import contextlib
import os
import unittest

import torch
import torch.distributed as dist
import torch.distributed._functional_collectives as _functional_collectives
from torch._dynamo.testing import CompileCounterWithBackend
from torch._subclasses.fake_tensor import FakeTensorMode
from torch.distributed.tensor import (
    DeviceMesh,
    distribute_tensor,
    DTensor,
    Partial,
    Replicate,
    Shard,
)
from torch.distributed.tensor._dtensor_spec import ShardOrderEntry
from torch.fx.experimental.proxy_tensor import make_fx
from torch.testing._internal.common_distributed import (
    MultiProcessTestCase,
    requires_nccl,
    skip_if_lt_x_gpu,
)
from torch.testing._internal.common_utils import (
    instantiate_parametrized_tests,
    parametrize,
    requires_cuda,
    run_tests,
    TestCase,
)
from torch.testing._internal.distributed.fake_pg import FakeStore
from torch.testing._internal.inductor_utils import GPU_TYPE, HAS_GPU
from torch.utils._debug_mode import (
    _OpCall,
    _RedistributeCall,
    _TritonKernelCall,
    DebugMode,
    hash_tensor_fn,
    norm_hash_fn,
)
from torch.utils._python_dispatch import TorchDispatchMode
from torch.utils._triton import has_triton_package


@requires_cuda
class TestDTensorDebugMode(TestCase):
    def tearDown(self):
        super().tearDown()
        dist.destroy_process_group()

    def setUp(self):
        super().setUp()
        self.world_size = 8
        store = FakeStore()
        dist.init_process_group(
            backend="fake", rank=0, world_size=self.world_size, store=store
        )
        self.device_type = "cuda"

    def test_debug_mode_mm(self):
        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        x = torch.randn(1, 8, requires_grad=False)
        y = torch.randn(1, 32, requires_grad=True)
        x_dtensor = DTensor.from_local(x, mesh, [Shard(0)], run_check=False)
        y_dtensor = DTensor.from_local(y, mesh, [Shard(0)], run_check=False)

        with DebugMode(record_torchfunction=True, record_ids=True) as debug_mode:
            torch.mm(x_dtensor, y_dtensor).sum()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  torch.mm(dt$0: f32[8, 8]| S(0), dt$1: f32[8, 32]| S(0))  ->  dt$7: f32[8, 32]| S(0)
    aten::mm(dt$0: f32[8, 8]| S(0), dt$1: f32[8, 32]| S(0))
      -> output: S(0)
      redistribute_input [implicit] (1, S(0) -> R)
        redistribute_input(t$2: f32[1, 32], trace: S(0)->R)
          _c10d_functional::all_gather_into_tensor(t$2: f32[1, 32], 8, 0)  ->  t$3: f32[8, 32]
          _c10d_functional::_wrap_tensor_autograd(t$3: f32[8, 32])  ->  t$4: f32[8, 32]
          _c10d_functional::wait_tensor(t$3: f32[8, 32])  ->  t$3: f32[8, 32]
      aten::mm(t$5: f32[1, 8], t$3: f32[8, 32])  ->  t$6: f32[1, 32]
  <method 'sum' of 'torch._C.TensorBase' objects>(dt$7: f32[8, 32]| S(0))  ->  dt$9: f32[]| P(sum)
    aten::sum(dt$7: f32[8, 32]| S(0))
      aten::sum(t$6: f32[1, 32])  ->  t$8: f32[]""",
        )

        self.assertTrue(isinstance(debug_mode.operators[0], _OpCall))
        self.assertTrue(isinstance(debug_mode.operators[3], _RedistributeCall))
        self.assertEqual(next(iter(debug_mode.operators[1])), torch.ops.aten.mm.default)

        # check stringification
        self.assertTrue(hasattr(debug_mode.operators[0], "args_str"))
        self.assertFalse(hasattr(debug_mode.operators[0], "args"))

        # check recording hook
        def mm(x, y):
            return (x @ y).sum()

        eager_out = mm(x_dtensor, y_dtensor)

        # check recording hook for compiled variant
        with (
            DebugMode() as debug_mode,
            DebugMode.record_outputs(),
            DebugMode.log_tensor_hashes(),
        ):
            compiled_out = torch.compile(mm, backend="aot_eager")(x_dtensor, y_dtensor)

        # check numerical equivalence
        self.assertTrue(torch.equal(eager_out, compiled_out))
        sum_op = next(
            iter(
                op
                for op in debug_mode.operators
                if isinstance(op, _OpCall) and str(op.op) == "aten.sum.default"
            )
        )
        self.assertTrue(torch.equal(sum_op.record["output"], eager_out.to_local()))
        self.assertTrue(
            "aten::sum(t: f32[1, 32])  ->  t: f32[]  # {'hash': "
            in debug_mode.debug_string()
        )

        # check tuple hash functions
        with (
            DebugMode() as debug_mode,
            DebugMode.log_tensor_hashes(hash_fn=["norm", "hash_tensor"]),
        ):
            mm(x_dtensor, y_dtensor)

        output_hash = debug_mode.operators[-1].log["hash"]
        norm_ = lambda x: norm_hash_fn(x, use_scalar=True)  # noqa: E731
        hash_ = lambda x: hash_tensor_fn(x, use_scalar=True)  # noqa: E731

        self.assertEqual(output_hash[0], norm_(eager_out))
        self.assertEqual(output_hash[1], hash_(eager_out))

        # some edge cases
        self.assertEqual(norm_(torch.tensor(torch.nan)), torch.nan)
        self.assertEqual(norm_(torch.tensor(torch.inf)), torch.inf)
        self.assertEqual(norm_(torch.complex(torch.ones(4), torch.zeros(4))), 4)
        self.assertEqual(hash_(torch.ones(4, dtype=torch.float8_e5m2)), 0)
        self.assertEqual(hash_(torch.ones(4, dtype=torch.int8)), 0)
        self.assertEqual(hash_(torch.ones(5, dtype=torch.int8)), 1)

    def test_debug_string_inside_context(self):
        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        x = torch.randn(1, 8, requires_grad=False)
        y = torch.randn(1, 32, requires_grad=True)
        x_dtensor = DTensor.from_local(x, mesh, [Shard(0)], run_check=False)
        y_dtensor = DTensor.from_local(y, mesh, [Shard(0)], run_check=False)

        with DebugMode() as debug_mode:
            torch.mm(x_dtensor, y_dtensor).sum()
            s0 = debug_mode.debug_string()
        s1 = debug_mode.debug_string()
        self.assertEqual(s0, s1)

    def test_debug_mode_backward(self):
        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        x = torch.randn(1, 8, requires_grad=True)
        y = torch.randn(8, 1, requires_grad=True)
        x_dtensor = DTensor.from_local(x, mesh, [Shard(0)], run_check=False)
        y_dtensor = DTensor.from_local(y, mesh, [Shard(1)], run_check=False)

        with DebugMode(
            record_torchfunction=True, record_stack_trace=True, record_output=False
        ) as debug_mode:
            z = x_dtensor + y_dtensor
            z.sum().backward()

        debug_string = debug_mode.debug_string(show_stack_trace=False)

        # Certain operations (clone, detach) appear in non-deterministic order during
        # backward pass redistribution.
        # For operations that appear multiple times, we include full parameters to
        # disambiguate (e.g., two different aten::add operations). For unique operations,
        # we use partial matching to avoid brittleness if the exact signature format changes.

        # Check for key forward operations
        self.assertIn(
            "<method 'add' of 'torch._C.TensorBase' objects>", debug_string
        )  # Unique - no params needed
        self.assertIn(
            "aten::add.Tensor(dt: f32[8, 8]| S(0), dt: f32[8, 8]| S(1))", debug_string
        )
        self.assertIn("redistribute_input [implicit] (1, S(1) -> S(0))", debug_string)
        self.assertIn(
            "redistribute_input(t: f32[8, 1], trace: S(1)->S(0))", debug_string
        )
        self.assertIn(
            "_dtensor::shard_dim_alltoall", debug_string
        )  # Unique - no params needed
        self.assertIn("aten::add.Tensor(t: f32[1, 8], t: f32[1, 8])", debug_string)

        # Check for sum operations
        self.assertIn(
            "<method 'sum' of 'torch._C.TensorBase' objects>", debug_string
        )  # Unique - no params needed
        self.assertIn("aten::sum(dt: f32[8, 8]| S(0))", debug_string)
        self.assertIn("aten::sum(t: f32[1, 8])", debug_string)

        # Check for backward operations
        self.assertIn(
            "torch._tensor.backward", debug_string
        )  # Unique - no params needed
        self.assertIn("aten::ones_like(dt: f32[]| P(sum)", debug_string)
        self.assertIn("aten::ones_like(t: f32[]", debug_string)
        self.assertIn("aten::expand(dt: f32[]| R, [8, 8])", debug_string)
        self.assertIn("aten::expand(t: f32[], [8, 8])", debug_string)

        # Check for redistribute operations in backward pass
        self.assertIn("redistribute_input(t: f32[8, 8], trace: R->S(1))", debug_string)
        self.assertIn("redistribute_input(t: f32[8, 8], trace: R->S(0))", debug_string)

        # Check for split operations in backward pass
        self.assertIn("aten::split.Tensor(t: f32[8, 8], 1, 1)", debug_string)
        self.assertIn("aten::split.Tensor(t: f32[8, 8], 1)", debug_string)

        # Check for _to_copy operations
        self.assertIn(
            "aten::_to_copy(t: f32[8, 1], dtype=torch.float32, layout=torch.strided, device=cpu)",
            debug_string,
        )
        self.assertIn(
            "aten::_to_copy(t: f32[1, 8], dtype=torch.float32, layout=torch.strided, device=cpu)",
            debug_string,
        )

        # Check that both clone and detach operations appear (order-independent)
        self.assertIn("aten::clone(t: f32[8, 1])", debug_string)
        self.assertIn("aten::clone(t: f32[1, 8])", debug_string)
        self.assertIn("aten::detach(t: f32[8, 1])", debug_string)
        self.assertIn("aten::detach(t: f32[1, 8])", debug_string)

        # check stack trace
        self.assertTrue("z.sum().backward()" in debug_mode.operators[-1].stack_trace)

    def test_stack_trace_in_compiled_region(self):
        class Foo(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(8, 4)
                self.l2 = torch.nn.Linear(4, 8)

            def forward(self, x):
                x = x + 2
                x = self.l1(x)
                x = x.relu()
                x = self.l2(x)
                x = x.sum()
                return x

        x = torch.randn(16, 8)
        model = torch.compile(Foo(), backend="aot_eager", fullgraph=True)

        # test forward nodes
        with DebugMode(
            record_stack_trace=True, run_compile_with_interpreter=True
        ) as debug_mode:
            out = model(x)

        op_calls = [op for op in debug_mode.operators if isinstance(op, _OpCall)]
        self.assertTrue("x = x + 2" in op_calls[0].stack_trace)
        self.assertTrue("x = self.l1(x)" in op_calls[1].stack_trace)
        self.assertTrue("x = x.relu()" in op_calls[3].stack_trace)
        self.assertTrue("x = x.sum()" in op_calls[-1].stack_trace)

        # test backward nodes
        with DebugMode(
            record_stack_trace=True, run_compile_with_interpreter=True
        ) as debug_mode:
            out.backward()

        op_calls = [op for op in debug_mode.operators if isinstance(op, _OpCall)]
        self.assertTrue("out.backward()" in op_calls[0].stack_trace)
        self.assertTrue("x = x.sum()" in op_calls[1].stack_trace)
        self.assertTrue("x = self.l2(x)" in op_calls[2].stack_trace)
        self.assertTrue("x = x.relu()" in op_calls[12].stack_trace)

    def test_debug_mode_densor_redistribution_trace(self):
        mesh = DeviceMesh(self.device_type, torch.arange(self.world_size).view(4, 2))

        x = torch.randn(16, 8, requires_grad=True)
        y = torch.randn(8, 16, requires_grad=True)
        x_dtensor = DTensor.from_local(x, mesh, [Shard(0), Shard(0)], run_check=False)
        y_dtensor = DTensor.from_local(y, mesh, [Shard(1), Shard(1)], run_check=False)
        x_dtensor._spec.shard_order = (ShardOrderEntry(tensor_dim=0, mesh_dims=(0, 1)),)
        y_dtensor._spec.shard_order = (ShardOrderEntry(tensor_dim=1, mesh_dims=(0, 1)),)
        with DebugMode(record_torchfunction=False, record_output=False) as debug_mode:
            torch.mm(x_dtensor, y_dtensor).sum()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  aten::mm(dt: f32[128, 8]| S(0)[0]S(0)[1], dt: f32[8, 128]| S(1)[0]S(1)[1])
    redistribute_input [implicit] (1, S(1)[0]S(1)[1] -> RR)
      redistribute_input(t: f32[8, 16], trace: S(1)[0]S(1)[1]->S(1)R->RR)
        _c10d_functional::all_gather_into_tensor(t: f32[8, 16], 2, 3)
        _c10d_functional::_wrap_tensor_autograd(t: f32[16, 16])
        _c10d_functional::wait_tensor(t: f32[16, 16])
        aten::chunk(t: f32[16, 16], 2)
        aten::cat(['t: f32[8, 16]', 't: f32[8, 16]'], 1)
        _c10d_functional::all_gather_into_tensor(t: f32[8, 32], 4, 1)
        _c10d_functional::_wrap_tensor_autograd(t: f32[32, 32])
        _c10d_functional::wait_tensor(t: f32[32, 32])
        aten::chunk(t: f32[32, 32], 4)
        aten::cat(['t: f32[8, 32]', 't: f32[8, 32]', 't: f32[8, 32]', 't: f32[8, 32]'], 1)
    aten::mm(t: f32[16, 8], t: f32[8, 128])
  aten::sum(dt: f32[128, 128]| S(0)[0]S(0)[1])
    aten::sum(t: f32[16, 128])""",
        )

    def test_debug_mode_explicit_redistribute(self):
        """Test that explicit user-called redistribute shows [explicit] annotation."""
        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        x = torch.randn(1, 8)
        x_dtensor = DTensor.from_local(x, mesh, [Shard(0)], run_check=False)

        with DebugMode() as debug_mode:
            x_dtensor.redistribute(mesh, [Replicate()])

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
    redistribute_input [explicit] (t: f32[1, 8], trace: S(0)->R)
      _c10d_functional::all_gather_into_tensor(t: f32[1, 8], 8, 0)  ->  t: f32[8, 8]
      _c10d_functional::_wrap_tensor_autograd(t: f32[8, 8])  ->  t: f32[8, 8]
      _c10d_functional::wait_tensor(t: f32[8, 8])  ->  t: f32[8, 8]""",
        )

        with DebugMode() as debug_mode:
            x_dtensor.full_tensor()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
    redistribute_input [explicit] (t: f32[1, 8], trace: S(0)->R)
      _c10d_functional::all_gather_into_tensor(t: f32[1, 8], 8, 0)  ->  t: f32[8, 8]
      _c10d_functional::_wrap_tensor_autograd(t: f32[8, 8])  ->  t: f32[8, 8]
      _c10d_functional::wait_tensor(t: f32[8, 8])  ->  t: f32[8, 8]
    aten::view(t: f32[8, 8], [8, 8])  ->  t: f32[8, 8]""",
        )

    def test_output_placements(self):
        """Test that output placements are recorded for multi-output DTensor ops."""
        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        # Use topk which returns multiple outputs (values, indices)
        # Shard on dim=1 (the topk dimension) to trigger redistribution
        x = torch.randn(8, 2, requires_grad=False)
        x_dtensor = DTensor.from_local(x, mesh, [Shard(1)], run_check=False)

        # Test with redistribution (slow path) - output placements should be recorded
        with DebugMode(record_output=True) as debug_mode:
            torch.topk(x_dtensor, k=4, dim=1)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  aten::topk(dt: f32[8, 16]| S(1), 4, 1)
    -> output: ('R', 'R')
    redistribute_input [implicit] (0, S(1) -> R)
      redistribute_input(t: f32[8, 2], trace: S(1)->R)
        _c10d_functional::all_gather_into_tensor(t: f32[8, 2], 8, 0)  ->  t: f32[64, 2]
        _c10d_functional::_wrap_tensor_autograd(t: f32[64, 2])  ->  t: f32[64, 2]
        _c10d_functional::wait_tensor(t: f32[64, 2])  ->  t: f32[64, 2]
        aten::chunk(t: f32[64, 2], 8)  ->  ['t: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]']
        aten::cat(['t: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]', 't: f32[8, 2]'], 1)  ->  t: f32[8, 16]
    aten::topk(t: f32[8, 16], 4, 1)  ->  ('t: f32[8, 4]', 't: i64[8, 4]')""",  # noqa: B950
        )

    def test_debug_mode_einsum(self):
        mesh = DeviceMesh(self.device_type, torch.arange(self.world_size).view(4, 2))

        # Create test tensors with mixed Partial placements: P(sum)R and RP(sum).
        # Per-input linearity allows bmm to operate directly on Partial inputs
        # without redistribution, producing P(sum)P(sum) output.
        # Numerics are verified in test_matrix_ops.py::test_mm_partial_inputs.
        a = torch.randn(16, 6, 8)
        b = torch.randn(8, 4, 4)

        a_dt = DTensor.from_local(a, mesh, [Partial(), Replicate()], run_check=False)
        b_dt = DTensor.from_local(b, mesh, [Replicate(), Partial()], run_check=False)

        # Capture the operator decomposition
        with DebugMode(record_torchfunction=True, record_output=False) as debug_mode:
            torch.einsum("bld,dnh->blnh", a_dt, b_dt)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  torch.functional.einsum(bld,dnh->blnh, dt: f32[16, 6, 8]| P(sum)R, dt: f32[8, 4, 4]| RP(sum))
    aten::unsqueeze(dt: f32[16, 6, 8]| P(sum)R, 3)
      aten::unsqueeze(t: f32[16, 6, 8], 3)
    aten::unsqueeze(dt: f32[16, 6, 8, 1]| P(sum)R, 4)
      aten::unsqueeze(t: f32[16, 6, 8, 1], 4)
    aten::permute(dt: f32[16, 6, 8, 1, 1]| P(sum)R, [0, 1, 3, 4, 2])
      aten::permute(t: f32[16, 6, 8, 1, 1], [0, 1, 3, 4, 2])
    aten::unsqueeze(dt: f32[8, 4, 4]| RP(sum), 3)
      aten::unsqueeze(t: f32[8, 4, 4], 3)
    aten::unsqueeze(dt: f32[8, 4, 4, 1]| RP(sum), 4)
      aten::unsqueeze(t: f32[8, 4, 4, 1], 4)
    aten::permute(dt: f32[8, 4, 4, 1, 1]| RP(sum), [3, 4, 1, 2, 0])
      aten::permute(t: f32[8, 4, 4, 1, 1], [3, 4, 1, 2, 0])
    aten::permute(dt: f32[16, 6, 1, 1, 8]| P(sum)R, [0, 1, 4, 2, 3])
      aten::permute(t: f32[16, 6, 1, 1, 8], [0, 1, 4, 2, 3])
    aten::view(dt: f32[16, 6, 8, 1, 1]| P(sum)R, [1, 96, 8])
      aten::view(t: f32[16, 6, 8, 1, 1], [1, 96, 8])
    aten::permute(dt: f32[1, 1, 4, 4, 8]| RP(sum), [4, 2, 3, 0, 1])
      aten::permute(t: f32[1, 1, 4, 4, 8], [4, 2, 3, 0, 1])
    aten::view(dt: f32[8, 4, 4, 1, 1]| RP(sum), [1, 8, 16])
      aten::view(t: f32[8, 4, 4, 1, 1], [1, 8, 16])
    aten::bmm(dt: f32[1, 96, 8]| P(sum)R, dt: f32[1, 8, 16]| RP(sum))
      aten::bmm(t: f32[1, 96, 8], t: f32[1, 8, 16])
    aten::view(dt: f32[1, 96, 16]| P(sum)P(sum), [16, 6, 1, 4, 4])
      aten::view(t: f32[1, 96, 16], [16, 6, 1, 4, 4])
    aten::permute(dt: f32[16, 6, 1, 4, 4]| P(sum)P(sum), [0, 1, 3, 4, 2])
      aten::permute(t: f32[16, 6, 1, 4, 4], [0, 1, 3, 4, 2])
    aten::view(dt: f32[16, 6, 4, 4, 1]| P(sum)P(sum), [16, 6, 4, 4])
      aten::view(t: f32[16, 6, 4, 4, 1], [16, 6, 4, 4])""",
        )

    def test_real_tensor(self):
        x = torch.randn(8, 8, 8)
        linear = torch.nn.Linear(8, 8)

        with DebugMode(record_torchfunction=True, record_output=False) as debug_mode:
            linear(x).sum()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  torch._C._nn.linear(t: f32[8, 8, 8], t: f32[8, 8], t: f32[8])
      aten::view(t: f32[8, 8, 8], [64, 8])
      aten::t(t: f32[8, 8])
      aten::addmm(t: f32[8], t: f32[64, 8], t: f32[8, 8])
      aten::view(t: f32[64, 8], [8, 8, 8])
  <method 'sum' of 'torch._C.TensorBase' objects>(t: f32[8, 8, 8])
      aten::sum(t: f32[8, 8, 8])""",
        )

    def test_fake_tensor(self):
        with FakeTensorMode():
            x = torch.randn(8, 8)
            y = torch.randn(8, 8, 8)

        with DebugMode(
            record_torchfunction=True, record_faketensor=True, record_output=False
        ) as debug_mode:
            torch.matmul(y, x)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  torch.matmul(ft: f32[8, 8, 8], ft: f32[8, 8])
      aten::view(ft: f32[8, 8, 8], [64, 8])
      aten::mm(ft: f32[64, 8], ft: f32[8, 8])
      aten::_unsafe_view(ft: f32[64, 8], [8, 8, 8])""",
        )

    def test_tensor_attributes(self):
        x = torch.randn(8, 8)
        x.a1 = "x1"
        x.a2 = "x2"
        y = torch.randn(8, 8, 8)
        y.a1 = "y"

        with DebugMode(
            record_torchfunction=True,
            record_faketensor=True,
            record_tensor_attributes=["a1", "a2"],
            store_original_args=True,
            record_output=False,
        ) as debug_mode:
            torch.matmul(y, x)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  torch.matmul(t: f32[8, 8, 8]{a1=y}, t: f32[8, 8]{a1=x1, a2=x2})
      aten::view(t: f32[8, 8, 8]{a1=y}, [64, 8])
      aten::mm(t: f32[64, 8], t: f32[8, 8]{a1=x1, a2=x2})
      aten::_unsafe_view(t: f32[64, 8], [8, 8, 8])""",
        )

        self.assertTrue(hasattr(debug_mode.operators[0], "args"))
        self.assertEqual(id(debug_mode.operators[0].args[0]), id(y))

    @parametrize("has_inner_mode", [True, False])
    @parametrize("has_outer_mode", [True, False])
    def test_nested_debug_mode(self, has_inner_mode, has_outer_mode):
        class DummyTorchDispatchMode1(TorchDispatchMode):
            def __torch_dispatch__(self, func, types, args=(), kwargs=None):
                return func(*args, **kwargs)

        class DummyTorchDispatchMode2(TorchDispatchMode):
            def __torch_dispatch__(self, func, types, args=(), kwargs=None):
                return func(*args, **kwargs)

        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        x = torch.randn(1, 8, requires_grad=True)
        y = torch.randn(1, 32, requires_grad=True)
        x_dtensor = DTensor.from_local(x, mesh, [Shard(0)], run_check=False)
        y_dtensor = DTensor.from_local(y, mesh, [Shard(0)], run_check=False)

        inner_mode = (
            DummyTorchDispatchMode1() if has_inner_mode else contextlib.nullcontext()
        )
        outer_mode = (
            DummyTorchDispatchMode2() if has_outer_mode else contextlib.nullcontext()
        )

        with outer_mode:
            with DebugMode() as debug_mode:
                with inner_mode:
                    torch.mm(x_dtensor, y_dtensor)

        self.assertTrue(
            "redistribute_input [implicit] (1, S(0) -> R)" in debug_mode.debug_string()
        )

    def test_debug_mode_higher_order_cond(self):
        """Test DebugMode with higher order operation."""
        x = torch.randn(1, 8, requires_grad=True)

        with DebugMode(record_torchfunction=True) as debug_mode:
            # rewrite torch.conda as torch.ops.higher_order.cond to avoid compilation
            torch.ops.higher_order.cond(
                torch.tensor(True), lambda x: x + 1, lambda x: x - 1, (x,)
            )

        # Verify that cond operations are captured in debug mode
        self.assertIn("torch.ops.higher_order.cond", debug_mode.debug_string())

    def test_compile(self):
        cnt = CompileCounterWithBackend("inductor")

        @torch.compile(backend=cnt)
        def f(x):
            return x.sin().cos()

        x = torch.randn(8)
        f(x)
        with DebugMode() as debug_mode:
            f(x)
            self.assertEqual(len(debug_mode.debug_string()), 0)
            f(x)
        f(x)
        self.assertEqual(
            cnt.frame_count, 1
        )  # check DebugMode doesn't trigger additional recompilations

    def test_annotate(self):
        x = torch.randn(8, 8)

        class Foo(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(8, 8)

            def forward(self, x):
                DebugMode._annotate("Foo")
                return self.l1(x)

        mod = Foo()
        with DebugMode(record_nn_module=True) as debug_mode:
            DebugMode._annotate("forward")
            mod(x)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  [annotate] forward
    [nn.Mod] Foo
    [annotate] Foo
      [nn.Mod] Foo.l1
        aten::t(t: f32[8, 8])  ->  t: f32[8, 8]
        aten::addmm(t: f32[8], t: f32[8, 8], t: f32[8, 8])  ->  t: f32[8, 8]""",
        )

        for backend in ["aot_eager", "inductor"]:
            with DebugMode(run_compile_with_interpreter=True) as debug_mode:
                torch.compile(mod, backend=backend, fullgraph=True)(x)

            if backend == "inductor":
                self.assertExpectedInline(
                    debug_mode.debug_string(),
                    """    aten::addmm.out(t: f32[8], t: f32[8, 8], t: f32[8, 8], out=t: f32[8, 8])  ->  t: f32[8, 8]""",
                )
            else:
                self.assertExpectedInline(
                    debug_mode.debug_string(),
                    """\
  [aot_eager region (compile)] enter
  [annotate] Foo
    aten::t(t: f32[8, 8])  ->  t: f32[8, 8]
    aten::addmm(t: f32[8], t: f32[8, 8], t: f32[8, 8])  ->  t: f32[8, 8]
  [aot_eager region (compile)] exit""",
                )

    def test_nn_module_in_eager(self):
        class Foo(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(4, 4)
                self.l2 = torch.nn.Linear(4, 4)

            def forward(self, x):
                return self.l2(self.l1(x))

        class Bar(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.abc = Foo()
                self.xyz = torch.nn.Linear(4, 4)

            def forward(self, x):
                return self.xyz(self.abc(x))

        mod = Bar()
        inp = torch.randn(4, 4)
        with DebugMode(record_nn_module=True, record_output=False) as debug_mode:
            _ = mod(inp)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
    [nn.Mod] Bar
      [nn.Mod] Bar.abc
        [nn.Mod] Bar.abc.l1
          aten::t(t: f32[4, 4])
          aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])
        [nn.Mod] Bar.abc.l2
          aten::t(t: f32[4, 4])
          aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])
      [nn.Mod] Bar.xyz
        aten::t(t: f32[4, 4])
        aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])""",
        )

        with DebugMode(record_stack_trace=True) as debug_mode:
            out = mod(inp).sum()
            out.backward()

        sum_op = [
            op
            for op in debug_mode.operators
            if isinstance(op, _OpCall) and str(op.op) == "aten.sum.dim_IntList"
        ][-1]
        self.assertTrue("self.l2(self.l1(x))" in sum_op.fwd_stack_trace)
        self.assertTrue(
            "self.l2(self.l1(x))" in debug_mode.debug_string(show_stack_trace=True)
        )

        # check that nn_module doesn't graph break in compiled regions
        fn = torch.compile(mod, backend="eager", fullgraph=True)
        with DebugMode(record_nn_module=True) as debug_mode:
            fn(inp)

    @torch._functorch.config.patch(guess_tangent_strides_as_outputs=True)
    def test_nn_module_in_compiled_regions(self):
        class Foo(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(4, 4)
                self.l2 = torch.nn.Linear(4, 4)

            def forward(self, x):
                return self.l2(self.l1(x).relu())

        class Bar(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.l3 = torch.nn.Linear(4, 4)

            def forward(self, x):
                return self.l3(x + 2.0)

        class Baz(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.foo = Foo()
                self.bar = Bar()

            def forward(self, x):
                return self.bar(self.foo(x))

        # Only region of module is compiled, test nn.Mod call hierarchy
        mod = Baz()
        mod.foo = torch.compile(mod.foo, backend="aot_eager", fullgraph=True)
        inp = torch.randn(4, 4)
        with DebugMode(
            record_nn_module=True, run_compile_with_interpreter=True
        ) as debug_mode:
            mod(inp).sum()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
    [nn.Mod] Baz
    [aot_eager region (compile)] enter
      [nn.Mod (compile)] L['self'].l1
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
      aten::relu(t: f32[4, 4])  ->  t: f32[4, 4]
      [nn.Mod (compile)] L['self'].l2
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
    [aot_eager region (compile)] exit
      [nn.Mod] Baz.bar
        aten::add.Tensor(t: f32[4, 4], 2.0)  ->  t: f32[4, 4]
        [nn.Mod] Baz.bar.l3
          aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
          aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
    aten::sum(t: f32[4, 4])  ->  t: f32[]""",
        )

        # Entire region is aot-eager compiled, with backwards
        mod = torch.compile(Baz(), backend="aot_eager", fullgraph=True)
        inp = torch.randn(4, 4)
        with DebugMode(
            record_nn_module=True, run_compile_with_interpreter=True
        ) as debug_mode:
            out = mod(inp).sum()
            out.backward()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  [aot_eager region (compile)] enter
    [nn.Mod (compile)] L['self'].foo
      [nn.Mod (compile)] L['self'].foo.l1
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
      aten::relu(t: f32[4, 4])  ->  t: f32[4, 4]
      [nn.Mod (compile)] L['self'].foo.l2
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
    [nn.Mod (compile)] L['self'].bar
      aten::add.Tensor(t: f32[4, 4], 2.0)  ->  t: f32[4, 4]
      [nn.Mod (compile)] L['self'].bar.l3
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::addmm(t: f32[4], t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
  [aot_eager region (compile)] exit
    aten::sum(t: f32[4, 4])  ->  t: f32[]
    aten::ones_like(t: f32[], pin_memory=False, memory_format=torch.preserve_format)  ->  t: f32[]
    aten::expand(t: f32[], [4, 4])  ->  t: f32[4, 4]
    aten::empty_strided([4, 4], [4, 1], dtype=torch.float32, layout=torch.strided, device=cpu, pin_memory=False)  ->  t: f32[4, 4]
    aten::copy_(t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
  [aot_eager region (compile)] enter
    [nn.Mod (compile)] L['self'].bar
      [nn.Mod (compile)] L['self'].bar.l3
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::mm(t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::mm(t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::sum.dim_IntList(t: f32[4, 4], [0], True)  ->  t: f32[1, 4]
        aten::view(t: f32[1, 4], [4])  ->  t: f32[4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
    [nn.Mod (compile)] L['self'].foo
      [nn.Mod (compile)] L['self'].foo.l2
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::mm(t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::mm(t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::sum.dim_IntList(t: f32[4, 4], [0], True)  ->  t: f32[1, 4]
        aten::view(t: f32[1, 4], [4])  ->  t: f32[4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
      aten::detach(t: f32[4, 4])  ->  t: f32[4, 4]
      aten::detach(t: f32[4, 4])  ->  t: f32[4, 4]
      aten::threshold_backward(t: f32[4, 4], t: f32[4, 4], 0)  ->  t: f32[4, 4]
      [nn.Mod (compile bwd)] L['self'].foo.l1
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::mm(t: f32[4, 4], t: f32[4, 4])  ->  t: f32[4, 4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
        aten::sum.dim_IntList(t: f32[4, 4], [0], True)  ->  t: f32[1, 4]
        aten::view(t: f32[1, 4], [4])  ->  t: f32[4]
        aten::t(t: f32[4, 4])  ->  t: f32[4, 4]
  [aot_eager region (compile)] exit
    aten::detach(t: f32[4, 4])  ->  t: f32[4, 4]
    aten::detach(t: f32[4, 4])  ->  t: f32[4, 4]
    aten::detach(t: f32[4])  ->  t: f32[4]
    aten::detach(t: f32[4, 4])  ->  t: f32[4, 4]
    aten::detach(t: f32[4])  ->  t: f32[4]
    aten::detach(t: f32[4])  ->  t: f32[4]""",
        )

    def test_record_function(self):
        def fn(x, y):
            z = x @ y
            with torch.profiler.record_function("this_is_ignored"):
                z = z + 1
            return z.sum()

        x = torch.randn(8, 4, requires_grad=True)
        y = torch.randn(4, 2, requires_grad=True)
        with DebugMode(run_compile_with_interpreter=True) as debug_mode:
            with torch.profiler.record_function("FWD"):
                out = torch.compile(fn, backend="aot_eager")(x, y)
            out.backward()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
  [record function] FWD
    [aot_eager region (compile)] enter
      aten::mm(t: f32[8, 4], t: f32[4, 2])  ->  t: f32[8, 2]
      aten::add.Tensor(t: f32[8, 2], 1)  ->  t: f32[8, 2]
      aten::sum(t: f32[8, 2])  ->  t: f32[]
      aten::t(t: f32[8, 4])  ->  t: f32[4, 8]
      aten::t(t: f32[4, 2])  ->  t: f32[2, 4]
    [aot_eager region (compile)] exit
      aten::detach(t: f32[4, 8])  ->  t: f32[4, 8]
      aten::detach(t: f32[2, 4])  ->  t: f32[2, 4]
    aten::ones_like(t: f32[], pin_memory=False, memory_format=torch.preserve_format)  ->  t: f32[]
  [aot_eager region (compile)] enter
    aten::expand(t: f32[], [8, 2])  ->  t: f32[8, 2]
    aten::mm(t: f32[4, 8], t: f32[8, 2])  ->  t: f32[4, 2]
    aten::mm(t: f32[8, 2], t: f32[2, 4])  ->  t: f32[8, 4]
  [aot_eager region (compile)] exit
    aten::detach(t: f32[8, 4])  ->  t: f32[8, 4]
    aten::detach(t: f32[4, 2])  ->  t: f32[4, 2]""",
        )

    def test_in_place_mutation(self):
        class Foo(torch.nn.Module):
            def forward(self, x):
                y = x + 1
                y.add_(1)
                return y

        mod = Foo()
        inp = torch.tensor(1)
        with (
            DebugMode(record_output=True) as debug_mode,
            DebugMode.log_tensor_hashes(hash_inputs=True),
        ):
            _ = mod(inp)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
    aten::add.Tensor(t: i64[], 1)  ->  t: i64[]  # {'input_hash': ((1.0, None), {}), 'hash': 2.0}
    aten::add_.Tensor(t: i64[], 1)  ->  t: i64[]  # {'input_hash': ((2.0, None), {}), 'hash': 3.0}""",
        )

    @unittest.skipIf(not HAS_GPU, "requires GPU")
    @unittest.skipIf(not has_triton_package(), "requires triton")
    def test_triton_kernel_logs(self):
        import triton

        from torch.testing._internal.triton_utils import add_kernel_autotuned

        def call_triton(x, y):
            output = torch.zeros_like(x)
            n_elements = output.numel()
            grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)  # noqa: E731
            add_kernel_autotuned[grid](x, y, output, n_elements)
            return output

        x = torch.randn(128, device=GPU_TYPE)
        y = torch.randn(128, device=GPU_TYPE)

        with DebugMode() as debug_mode:
            torch.compile(call_triton)(x, y)

        triton_calls = [
            op for op in debug_mode.operators if isinstance(op, _TritonKernelCall)
        ]
        self.assertGreater(len(triton_calls), 0)
        self.assertIn("[triton]", triton_calls[0].render([]))

    def test_check_hash_mismatches(self):
        x = torch.randn(64, 64, device=GPU_TYPE)
        x_different = torch.randn(64, 64, device=GPU_TYPE)

        # Identical runs should have no mismatches
        with DebugMode() as dm1, DebugMode.log_tensor_hashes():
            x.sin().sum()
        with DebugMode() as dm2, DebugMode.log_tensor_hashes():
            x.sin().sum()
        mismatches = DebugMode.check_hash_mismatches(dm1.logs, dm2.logs)
        self.assertEqual(len(mismatches), 0)

        # Different inputs should produce hash mismatches
        with DebugMode() as dm3, DebugMode.log_tensor_hashes():
            x_different.sin().sum()

        # Check that mismatches are detected
        mismatches = DebugMode.check_hash_mismatches(dm1.logs, dm3.logs)
        self.assertEqual(len(mismatches), 2)
        self.assertEqual(
            [call["call"] for call in mismatches], ["aten::sin", "aten::sum"]
        )

    @unittest.skipIf(
        not torch.cuda.is_available()
        or torch.cuda.get_device_properties(0).total_memory < 2**26,
        "Being conservative, test peak memory is 25MB?",
    )
    def test_tensor_hash_redistribute(self):
        # test that hashing collectives gives correct results
        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        local_tensor = torch.ones(2**18, device=self.device_type)
        dt = DTensor.from_local(local_tensor, mesh, [Shard(0)], run_check=False)

        with DebugMode() as debug_mode, DebugMode.log_tensor_hashes():
            dt.redistribute(mesh, [Replicate()])

        # Find all_gather hash
        all_gather_logs = [
            op
            for op in debug_mode.logs
            if isinstance(op, _OpCall)
            and op.op == torch.ops._c10d_functional.all_gather_into_tensor.default
        ]
        self.assertEqual(len(all_gather_logs), 1)
        actual_hash = all_gather_logs[0].log["hash"]
        self.assertEqual(actual_hash, float(local_tensor.numel() * self.world_size))

    @unittest.skipIf(not HAS_GPU, "requires GPU")
    @unittest.skipIf(not has_triton_package(), "requires triton")
    def test_check_triton_hash_mismatches(self):
        import triton

        from torch.testing._internal.triton_utils import add_kernel_autotuned

        def call_triton(x, y):
            output = torch.zeros_like(x)
            n_elements = output.numel()
            grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)  # noqa: E731
            add_kernel_autotuned[grid](x, y, output, n_elements)
            return output

        a = torch.randn(128, device=GPU_TYPE)
        b = torch.randn(128, device=GPU_TYPE)
        c = torch.randn(128, device=GPU_TYPE)

        # Run with hash logging to verify triton kernels can be hashed
        with DebugMode() as dm_t1, DebugMode.log_tensor_hashes(hash_inputs=True):
            torch.compile(call_triton)(a, b)

        # Different inputs should have different hashes in triton kernels
        with DebugMode() as dm_t2, DebugMode.log_tensor_hashes(hash_inputs=True):
            torch.compile(call_triton)(a, c)

        # Compare triton kernel hashes
        mismatches = DebugMode.check_hash_mismatches(
            dm_t1.logs, dm_t2.logs, compare_inputs=True
        )
        triton_mismatches = [m for m in mismatches if m["call_type"] == "triton kernel"]
        self.assertGreater(len(triton_mismatches), 0)

        # check both input & output hash mismatches are detected
        self.assertGreater(len([m for m in triton_mismatches if m["is_input_hash"]]), 0)
        self.assertGreater(
            len([m for m in triton_mismatches if not m["is_input_hash"]]), 0
        )

    def test_check_structure_mismatches(self):
        x = torch.randn(32, 32, device=self.device_type)

        with DebugMode() as dm1, DebugMode.log_tensor_hashes():
            x.sin()
        with DebugMode() as dm2, DebugMode.log_tensor_hashes():
            x.cos()
        with DebugMode() as dm3, DebugMode.log_tensor_hashes():
            x.sin().cos()

        with self.assertRaisesRegex(ValueError, "Operators don't match"):
            DebugMode.check_hash_mismatches(dm1.logs, dm2.logs)

        with self.assertRaisesRegex(ValueError, "Log lengths don't match"):
            DebugMode.check_hash_mismatches(dm1.logs, dm3.logs)

    def test_pretty_print_dtensor_make_fx(self):
        mesh = DeviceMesh(self.device_type, list(range(self.world_size)))

        A = torch.randn(8, 32)
        B = torch.randn(32, 32)
        dA = distribute_tensor(A, mesh, [Shard(0)]).requires_grad_()
        dB = distribute_tensor(B, mesh, [Replicate()]).requires_grad_()

        def f(dA, dB):
            dy = dA @ dB
            loss = dy.sum()
            loss.backward()
            return dA.grad, dB.grad

        # We actually need the tracing_mode='fake' here, or to trace under a FakeTensorMode.
        # make_fx has some logic to ensure we don't accidentally stash real tensors in the graph
        # so we won't stash our DTensors properly if they don't hold Fake inner tensors
        gm = make_fx(f, tracing_mode="fake")(dA, dB)
        # DCE isn't necessary here, there were just a lot of dead detach() nodes that spammed the graph
        gm.graph.eliminate_dead_code()
        gm.recompile()
        # Colored is nice for actual viewing, not using in this test though
        gm_str = gm.print_readable(colored=False, print_output=False)
        self.assertTrue('"DTensor(f32[8, 32], S(0))" = torch.ops.aten.mm' in gm_str)

    @torch._functorch.config.patch(guess_tangent_strides_as_outputs=True)
    def test_invoke_subgraph(self):
        # Test that DebugMode can trace the operations inside
        # invoke_subgraph HOP

        @torch.compiler.nested_compile_region
        def gn(x):
            a = torch.sin(x)
            return a

        def fn(x):
            foo1 = gn(x)
            y = foo1 * 2
            foo2 = gn(y)
            return foo2

        opt_fn = torch.compile(fn, backend="aot_eager", fullgraph=True)

        x = torch.randn(8, 8, requires_grad=True)
        x_clone = x.detach().clone().requires_grad_(True)
        x.grad = None
        x_clone.grad = None

        ref = fn(x)
        ref.sum().backward()

        with DebugMode() as debug_mode:
            res = opt_fn(x_clone)
            res.sum().backward()

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
    torch.ops.higher_order.invoke_subgraph(partitioned_fw_subgraph_0_0, t: f32[8, 8])  ->  ('t: f32[8, 8]', 't: f32[8, 8]')
    [annotate] [enter InvokeSubgraph HOP] partitioned_fw_subgraph_0_0
      aten::sin(t: f32[8, 8])  ->  t: f32[8, 8]
    [annotate] [exit InvokeSubgraph HOP] partitioned_fw_subgraph_0_0
    aten::mul.Tensor(t: f32[8, 8], 2)  ->  t: f32[8, 8]
    torch.ops.higher_order.invoke_subgraph(partitioned_fw_subgraph_0_0, t: f32[8, 8])  ->  ('t: f32[8, 8]', 't: f32[8, 8]')
    [annotate] [enter InvokeSubgraph HOP] partitioned_fw_subgraph_0_0
      aten::sin(t: f32[8, 8])  ->  t: f32[8, 8]
    [annotate] [exit InvokeSubgraph HOP] partitioned_fw_subgraph_0_0
    aten::sum(t: f32[8, 8])  ->  t: f32[]
    aten::ones_like(t: f32[], pin_memory=False, memory_format=torch.preserve_format)  ->  t: f32[]
    aten::expand(t: f32[], [8, 8])  ->  t: f32[8, 8]
    aten::empty_strided([8, 8], [8, 1], dtype=torch.float32, layout=torch.strided, device=cpu, pin_memory=False)  ->  t: f32[8, 8]
    aten::copy_(t: f32[8, 8], t: f32[8, 8])  ->  t: f32[8, 8]
    torch.ops.higher_order.invoke_subgraph(partitioned_bw_subgraph_0_0, t: f32[8, 8], t: f32[8, 8])  ->  ('t: f32[8, 8]',)
    [annotate] [enter InvokeSubgraph HOP] partitioned_bw_subgraph_0_0
      aten::cos(t: f32[8, 8])  ->  t: f32[8, 8]
      aten::mul.Tensor(t: f32[8, 8], t: f32[8, 8])  ->  t: f32[8, 8]
    [annotate] [exit InvokeSubgraph HOP] partitioned_bw_subgraph_0_0
    aten::mul.Tensor(t: f32[8, 8], 2)  ->  t: f32[8, 8]
    torch.ops.higher_order.invoke_subgraph(partitioned_bw_subgraph_0_0, t: f32[8, 8], t: f32[8, 8])  ->  ('t: f32[8, 8]',)
    [annotate] [enter InvokeSubgraph HOP] partitioned_bw_subgraph_0_0
      aten::cos(t: f32[8, 8])  ->  t: f32[8, 8]
      aten::mul.Tensor(t: f32[8, 8], t: f32[8, 8])  ->  t: f32[8, 8]
    [annotate] [exit InvokeSubgraph HOP] partitioned_bw_subgraph_0_0
    aten::detach(t: f32[8, 8])  ->  t: f32[8, 8]""",  # noqa: B950
            ignore_comments=True,
        )

        self.assertEqual(ref, res)
        self.assertEqual(x.grad, x_clone.grad)

    def test_nested_invoke_subgraph(self):
        # Test that DebugMode can trace the operations inside
        # invoke_subgraph HOP
        @torch.compiler.nested_compile_region
        def ggn(x):
            a = torch.cos(x)
            return a

        @torch.compiler.nested_compile_region
        def gn(x):
            x = x + 1
            x = ggn(x)
            a = torch.sin(x)
            return a

        def fn(x):
            foo1 = gn(x)
            y = foo1 * 2
            return y

        opt_fn = torch.compile(fn, backend="aot_eager", fullgraph=True)

        x = torch.randn(8, 8)

        ref = fn(x)

        with DebugMode() as debug_mode:
            res = opt_fn(x)
        self.assertEqual(ref, res)

        self.assertExpectedInline(
            debug_mode.debug_string(),
            """\
    torch.ops.higher_order.invoke_subgraph(subgraph_1, t: f32[8, 8])  ->  ('t: f32[8, 8]',)
    [annotate] [enter InvokeSubgraph HOP] subgraph_1
      aten::add.Tensor(t: f32[8, 8], 1)  ->  t: f32[8, 8]
      torch.ops.higher_order.invoke_subgraph(subgraph_0, t: f32[8, 8])  ->  ('t: f32[8, 8]',)
      [annotate] [enter InvokeSubgraph HOP] subgraph_0
        aten::cos(t: f32[8, 8])  ->  t: f32[8, 8]
      [annotate] [exit InvokeSubgraph HOP] subgraph_0
      aten::sin(t: f32[8, 8])  ->  t: f32[8, 8]
    [annotate] [exit InvokeSubgraph HOP] subgraph_1
    aten::mul.Tensor(t: f32[8, 8], 2)  ->  t: f32[8, 8]""",  # noqa: B950
            ignore_comments=True,
        )


class TestDebugModeUtils(TestCase):
    """Test DebugMode with NCCL backend without using DTensor."""

    def test_hash_empty_tensor(self):
        t = torch.tensor([])
        # hash tensor fn should not error out with empty tensor
        out = torch.utils._debug_mode.hash_tensor_fn(t)
        self.assertTrue(isinstance(out, torch.Tensor))
        out = torch.utils._debug_mode.hash_tensor_fn(t, use_scalar=True)
        self.assertTrue(isinstance(out, int))


class TestDTensorDebugModeNCCLBackend(MultiProcessTestCase):
    @property
    def world_size(self):
        return 2  # Need at least 2 ranks for collectives

    def setUp(self):
        super().setUp()
        self._spawn_processes()

    def _init_process_group(self):
        """Initialize NCCL process group for each spawned process."""
        torch.cuda.set_device(self.rank)
        store = dist.FileStore(self.file_name, self.world_size)
        dist.init_process_group(
            "nccl",
            world_size=self.world_size,
            rank=self.rank,
            store=store,
        )
        self.device = f"cuda:{self.rank}"

    def _destroy_process_group(self):
        """Destroy the process group."""
        dist.destroy_process_group()

    def tearDown(self):
        super().tearDown()
        try:
            os.remove(self.file_name)
        except OSError:
            pass

    @requires_nccl()
    @skip_if_lt_x_gpu(2)
    def test_allgather_base(self):
        self._init_process_group()
        tensor = torch.ones(10, 10, device=torch.device(self.device)) * (self.rank + 1)
        # Output size must be world_size * input_size
        output_tensor = torch.zeros(
            10 * self.world_size, 10, device=torch.device(self.device)
        )

        with DebugMode() as debug_mode, DebugMode.log_tensor_hashes(hash_inputs=True):
            dist.all_gather_into_tensor(output_tensor, tensor)

        self.assertTrue("c10d::_allgather_base_" in debug_mode.debug_string())

        hash_ = lambda x: norm_hash_fn(x, use_scalar=True)  # noqa: E731

        self.assertEqual(debug_mode.operators[-1].log["hash"][0], hash_(output_tensor))

        # Verify each rank's contribution
        for i in range(self.world_size):
            expected_slice = torch.ones(10, 10, device=self.device) * (i + 1)
            self.assertEqual(output_tensor[i * 10 : (i + 1) * 10], expected_slice)

        self._destroy_process_group()

    @requires_nccl()
    @skip_if_lt_x_gpu(2)
    def test_allgather_base_async_op(self):
        """Test all_gather_into_tensor with async_op=True."""
        self._init_process_group()
        tensor = torch.ones(10, 10, device=torch.device(self.device)) * (self.rank + 1)
        # Output size must be world_size * input_size
        output_tensor = torch.zeros(
            10 * self.world_size, 10, device=torch.device(self.device)
        )

        with DebugMode() as debug_mode, DebugMode.log_tensor_hashes(hash_inputs=True):
            # Call with async_op=True returns a work handle
            work = dist.all_gather_into_tensor(output_tensor, tensor, async_op=True)
            # Wait for the async operation to complete
            work.wait()

        self.assertTrue("c10d::_allgather_base_" in debug_mode.debug_string())
        hash_ = lambda x: norm_hash_fn(x, use_scalar=True)  # noqa: E731

        # TODO: the output hash not correct for async op, we should either not log the hash
        # or wait before logging.
        # https://github.com/pytorch/pytorch/pull/169295
        # self.assertEqual(debug_mode.operators[-1].log["hash"][0], hash_(output_tensor))
        self.assertEqual(
            debug_mode.operators[-1].log["input_hash"][0][1], hash_(tensor)
        )

        # Verify each rank's contribution
        for i in range(self.world_size):
            expected_slice = torch.ones(10, 10, device=self.device) * (i + 1)
            self.assertEqual(output_tensor[i * 10 : (i + 1) * 10], expected_slice)

        self._destroy_process_group()

    @requires_nccl()
    @skip_if_lt_x_gpu(2)
    def test_allgather_functional_with_async_collective_tensor(self):
        self._init_process_group()
        tensor = torch.ones(10, 10, device=torch.device(self.device)) * (self.rank + 1)

        # Use functional collectives which return AsyncCollectiveTensor
        with DebugMode() as debug_mode, DebugMode.log_tensor_hashes():
            result = _functional_collectives.all_gather_tensor(
                tensor, gather_dim=0, group=dist.group.WORLD
            )

        result = result.wait()
        hash_ = lambda x: norm_hash_fn(x, use_scalar=True)  # noqa: E731

        self.assertEqual(debug_mode.operators[-1].log["hash"], hash_(result))

        self.assertTrue(
            "_c10d_functional::all_gather_into_tensor" in debug_mode.debug_string()
        )

        # Verify the result shape - should be world_size times bigger
        self.assertEqual(result.shape[0], tensor.shape[0] * self.world_size)
        # Verify each rank's contribution
        for i in range(self.world_size):
            expected_slice = torch.ones(10, 10, device=self.device) * (i + 1)
            self.assertEqual(result[i * 10 : (i + 1) * 10], expected_slice)

        self._destroy_process_group()


instantiate_parametrized_tests(TestDTensorDebugMode)


if __name__ == "__main__":
    run_tests()



# Owner(s): ["module: sdpa"]
import unittest
from collections import namedtuple
from contextlib import contextmanager, nullcontext

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.attention import (
    activate_flash_attention_impl,
    restore_flash_attention_impl,
)
from torch.nn.attention.varlen import varlen_attn, varlen_attn_out
from torch.testing._internal.common_cuda import (
    IS_SM90,
    PLATFORM_SUPPORTS_CK_SDPA,
    PLATFORM_SUPPORTS_FLASH_ATTENTION,
    SM100OrLater,
    SM120OrLater,
    SM90OrLater,
)
from torch.testing._internal.common_device_type import instantiate_device_type_tests
from torch.testing._internal.common_nn import NNTestCase
from torch.testing._internal.common_utils import (
    decorateIf,
    parametrize,
    run_tests,
    TEST_WITH_ROCM,
)
from torch.utils._python_dispatch import TorchDispatchMode


@contextmanager
def use_fa3():
    try:
        activate_flash_attention_impl("FA3")
    except (ModuleNotFoundError, RuntimeError) as err:
        raise unittest.SkipTest(
            "FA3 backend not available (flash_attn_interface missing)"
        ) from err
    try:
        yield
    finally:
        restore_flash_attention_impl()


@contextmanager
def use_fa4():
    try:
        activate_flash_attention_impl("FA4")
    except (ModuleNotFoundError, RuntimeError) as err:
        raise unittest.SkipTest("FA4 backend not available") from err
    try:
        yield
    finally:
        restore_flash_attention_impl()


def _use_backend(backend):
    return {"fa2": nullcontext, "fa3": use_fa3, "fa4": use_fa4}[backend]()


def _varlen_backends(*, include_fa4_paged_kv: bool) -> list[str]:
    fa4_supported = (
        SM100OrLater if include_fa4_paged_kv else SM90OrLater
    ) and not SM120OrLater
    return ["fa2"] + (["fa3"] if IS_SM90 else []) + (["fa4"] if fa4_supported else [])


VarlenShape = namedtuple(
    "VarlenShape",
    ["batch_size", "max_seq_len", "embed_dim", "num_heads", "num_kv_heads"],
    defaults=[None],
)


class OpLoggingMode(TorchDispatchMode):
    """Logging mode that captures all dispatched operations"""

    def __init__(self):
        self.called_ops = []

    def __torch_dispatch__(self, func, types, args=(), kwargs=None):
        op_name = str(func)
        self.called_ops.append(op_name)
        return func(*args, **(kwargs or {}))


class AttentionBlock(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        device: torch.device,
        dtype: torch.dtype,
        num_kv_heads: int | None = None,
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.num_kv_heads = num_kv_heads if num_kv_heads is not None else num_heads
        self.head_dim = embed_dim // num_heads

        self.q_proj = nn.Linear(
            embed_dim,
            num_heads * self.head_dim,
            bias=False,
            device=device,
            dtype=dtype,
        )
        self.kv_proj = nn.Linear(
            embed_dim,
            2 * self.num_kv_heads * self.head_dim,
            bias=False,
            device=device,
            dtype=dtype,
        )
        self.out_proj = nn.Linear(
            embed_dim, embed_dim, bias=False, device=device, dtype=dtype
        )

    @property
    def enable_gqa(self):
        return self.num_kv_heads != self.num_heads

    def get_varlen_qkv(
        self,
        x_packed: torch.Tensor,
    ):
        q = self.q_proj(x_packed).view(-1, self.num_heads, self.head_dim)
        kv = self.kv_proj(x_packed).view(-1, 2, self.num_kv_heads, self.head_dim)
        k, v = kv[:, 0], kv[:, 1]
        return q, k, v

    def forward_varlen(
        self,
        x_packed: torch.Tensor,
        cu_seq: torch.Tensor,
        max_len: int,
        scale: float | None = None,
        window_size: tuple[int, int] = (-1, -1),
    ):
        q, k, v = self.get_varlen_qkv(x_packed)

        attn_out = varlen_attn(
            q,
            k,
            v,
            cu_seq,
            cu_seq,
            max_len,
            max_len,
            scale=scale,
            window_size=window_size,
            enable_gqa=self.enable_gqa,
        )
        attn_out = attn_out.view(-1, self.embed_dim)

        return self.out_proj(attn_out)

    def forward_sdpa(
        self,
        x_padded: torch.Tensor,
        seq_lengths: torch.Tensor,
        scale: float | None = None,
        window_size: tuple[int, int] = (-1, -1),
    ):
        batch_size, seq_len, _ = x_padded.shape

        q = self.q_proj(x_padded)
        kv = self.kv_proj(x_padded)
        k, v = kv.view(batch_size, seq_len, 2, self.num_kv_heads, self.head_dim).unbind(
            dim=2
        )

        padding_mask = (
            torch.arange(seq_len, device=x_padded.device)[None, :]
            < seq_lengths[:, None]
        )

        attn_mask = padding_mask[:, None, None, :].expand(
            batch_size, self.num_heads, seq_len, seq_len
        )

        if window_size == (-1, 0):
            causal_mask = torch.tril(
                torch.ones(seq_len, seq_len, device=x_padded.device, dtype=torch.bool)
            )
            # Combine: attention allowed where BOTH padding is valid AND causal constraint is met
            attn_mask = attn_mask & causal_mask[None, None, :, :]

        if window_size[0] >= 0 or window_size[1] >= 0:
            window_mask = torch.zeros(
                seq_len, seq_len, dtype=torch.bool, device=x_padded.device
            )
            for i in range(seq_len):
                start = i - window_size[0] if window_size[0] >= 0 else 0
                end = i + window_size[1] + 1 if window_size[1] >= 0 else seq_len
                start = max(start, 0)
                end = min(end, seq_len)
                window_mask[i, start:end] = True
            attn_mask = attn_mask & window_mask[None, None, :, :]

        q = q.view(batch_size, seq_len, self.num_heads, self.head_dim).transpose(1, 2)
        k = k.view(batch_size, seq_len, self.num_kv_heads, self.head_dim).transpose(
            1, 2
        )
        v = v.view(batch_size, seq_len, self.num_kv_heads, self.head_dim).transpose(
            1, 2
        )

        # Don't pass is_causal since we already incorporated it into attn_mask.
        if self.enable_gqa:
            # Force math backend for GQA
            with torch.nn.attention.sdpa_kernel(torch.nn.attention.SDPBackend.MATH):
                attn_out = F.scaled_dot_product_attention(
                    q,
                    k,
                    v,
                    attn_mask=attn_mask,
                    scale=scale,
                    enable_gqa=True,
                )
        else:
            attn_out = F.scaled_dot_product_attention(
                q,
                k,
                v,
                attn_mask=attn_mask,
                scale=scale,
            )

        attn_out = (
            attn_out.transpose(1, 2)
            .contiguous()
            .view(batch_size, seq_len, self.embed_dim)
        )

        return self.out_proj(attn_out)


def pack_sequences(seqs, device):
    x_packed = torch.cat(seqs, dim=0)
    seq_lens = torch.tensor([len(s) for s in seqs], device=device)
    cu_seq = torch.zeros(len(seqs) + 1, device=device, dtype=torch.int32)
    cu_seq[1:] = seq_lens.cumsum(0)
    max_len = seq_lens.max().item()

    return x_packed, cu_seq, max_len


def create_variable_length_batch(
    shape: VarlenShape, device: torch.device, dtype: torch.dtype
):
    seq_lengths = []
    for _ in range(shape.batch_size):
        length = torch.randint(1, shape.max_seq_len // 64 + 1, (1,)).item() * 64
        seq_lengths.append(min(length, shape.max_seq_len))

    sequences_fp32 = [
        torch.randn(seq_len, shape.embed_dim, device=device, dtype=torch.float32)
        for seq_len in seq_lengths
    ]

    x_packed_fp32, cu_seq, max_len = pack_sequences(sequences_fp32, device)
    seq_lengths = torch.tensor(seq_lengths, device=device)

    x_padded_fp32 = torch.zeros(
        shape.batch_size, max_len, shape.embed_dim, device=device, dtype=torch.float32
    )
    start_idx = 0
    for i, seq_len in enumerate(seq_lengths):
        end_idx = start_idx + seq_len
        x_padded_fp32[i, :seq_len] = x_packed_fp32[start_idx:end_idx]
        start_idx = end_idx

    x_packed = x_packed_fp32.detach().clone().to(dtype).requires_grad_(True)
    x_padded = x_padded_fp32.detach().clone().to(dtype).requires_grad_(True)
    x_padded_ref = x_padded_fp32.detach().clone().requires_grad_(True)

    return {
        "seq_lengths": seq_lengths,
        "cu_seq": cu_seq,
        "x_packed": x_packed,
        "x_padded": x_padded,
        "x_padded_ref": x_padded_ref,
        "max_len": max_len,
    }


class TestVarlenAttention(NNTestCase):
    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    @parametrize(
        "sdpa_backend",
        ["aotriton", "ck"] if PLATFORM_SUPPORTS_CK_SDPA else ["aotriton"],
    )
    @parametrize(
        "backend",
        _varlen_backends(include_fa4_paged_kv=False),
    )
    def test_basic_functionality(self, device, dtype, backend, sdpa_backend=None):
        if TEST_WITH_ROCM:
            torch.backends.cuda.preferred_rocm_fa_library(sdpa_backend)

        torch.manual_seed(42)

        shape = VarlenShape(batch_size=2, max_seq_len=512, embed_dim=1024, num_heads=16)

        attention_block = AttentionBlock(
            shape.embed_dim, shape.num_heads, device, dtype
        )

        total_tokens = shape.batch_size * shape.max_seq_len
        x_packed = torch.randn(
            total_tokens,
            shape.embed_dim,
            device=device,
            dtype=dtype,
            requires_grad=True,
        )
        cu_seq = torch.tensor(
            [0, shape.max_seq_len, total_tokens], device=device, dtype=torch.int32
        )

        with _use_backend(backend):
            output = attention_block.forward_varlen(x_packed, cu_seq, shape.max_seq_len)

            self.assertEqual(output.shape, (total_tokens, shape.embed_dim))
            self.assertEqual(output.device, torch.device(device))
            self.assertEqual(output.dtype, dtype)

            # varlen_attn_out should produce the same result and write into the buffer
            with torch.no_grad():
                q, k, v = attention_block.get_varlen_qkv(x_packed)
                expected = varlen_attn(
                    q, k, v, cu_seq, cu_seq, shape.max_seq_len, shape.max_seq_len
                )
                out_buf = torch.empty_like(expected)
                actual = varlen_attn_out(
                    out_buf,
                    q,
                    k,
                    v,
                    cu_seq,
                    cu_seq,
                    shape.max_seq_len,
                    shape.max_seq_len,
                )
                self.assertEqual(actual.data_ptr(), out_buf.data_ptr())
                self.assertEqual(out_buf, expected)

            varlen_grad_out = torch.ones_like(output)

            varlen_grad = torch.autograd.grad(
                outputs=output,
                inputs=x_packed,
                grad_outputs=varlen_grad_out,
                retain_graph=True,
                create_graph=False,
                allow_unused=False,
            )[0]

            self.assertIsNotNone(varlen_grad)
            self.assertEqual(varlen_grad.shape, x_packed.shape)
            self.assertEqual(varlen_grad.dtype, x_packed.dtype)

    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @parametrize(
        "sdpa_backend",
        ["aotriton", "ck"] if PLATFORM_SUPPORTS_CK_SDPA else ["aotriton"],
    )
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    def test_custom_op_compliance(self, device, dtype, sdpa_backend=None):
        if TEST_WITH_ROCM:
            torch.backends.cuda.preferred_rocm_fa_library(sdpa_backend)
        torch.manual_seed(42)

        shape = VarlenShape(batch_size=2, max_seq_len=512, embed_dim=1024, num_heads=16)

        attention_block = AttentionBlock(
            shape.embed_dim, shape.num_heads, device, dtype
        )

        total_tokens = shape.batch_size * shape.max_seq_len
        x_packed = torch.randn(
            total_tokens,
            shape.embed_dim,
            device=device,
            dtype=dtype,
        )
        cu_seq = torch.tensor(
            [0, shape.max_seq_len, total_tokens], device=device, dtype=torch.int32
        )

        q, k, v = attention_block.get_varlen_qkv(x_packed)

        torch.library.opcheck(
            torch.ops.torch_attn._varlen_attn,
            (q, k, v, cu_seq, cu_seq, shape.max_seq_len, shape.max_seq_len, False),
        )

        out, lse, rng_state = torch.ops.torch_attn._varlen_attn(
            q, k, v, cu_seq, cu_seq, shape.max_seq_len, shape.max_seq_len, False
        )
        grad_out = torch.randn_like(out)

        # we don't support double backward
        # skipping test_autograd_registration, test_aot_dispatch_dynamic, test_aot_dispatch_static
        torch.library.opcheck(
            torch.ops.torch_attn._varlen_attn_backward,
            (
                grad_out,
                q,
                k,
                v,
                out,
                lse,
                cu_seq,
                cu_seq,
                shape.max_seq_len,
                shape.max_seq_len,
                False,
                rng_state,
            ),
            test_utils=["test_schema", "test_faketensor"],
        )

        # opcheck for _varlen_attn_out (no backward)
        out_buf = torch.empty_like(q)
        torch.library.opcheck(
            torch.ops.torch_attn._varlen_attn_out,
            (
                out_buf,
                q,
                k,
                v,
                cu_seq,
                cu_seq,
                shape.max_seq_len,
                shape.max_seq_len,
                False,
            ),
            test_utils=["test_schema", "test_faketensor"],
        )

    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @parametrize(
        "sdpa_backend",
        ["aotriton", "ck"] if PLATFORM_SUPPORTS_CK_SDPA else ["aotriton"],
    )
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    def test_custom_op_registration(self, device, dtype, sdpa_backend=None):
        if TEST_WITH_ROCM:
            torch.backends.cuda.preferred_rocm_fa_library(sdpa_backend)
        torch.manual_seed(42)

        shape = VarlenShape(batch_size=2, max_seq_len=512, embed_dim=1024, num_heads=16)

        attention_block = AttentionBlock(
            shape.embed_dim, shape.num_heads, device, dtype
        )

        total_tokens = shape.batch_size * shape.max_seq_len
        x_packed = torch.randn(
            total_tokens,
            shape.embed_dim,
            device=device,
            dtype=dtype,
            requires_grad=True,
        )
        cu_seq = torch.tensor(
            [0, shape.max_seq_len, total_tokens], device=device, dtype=torch.int32
        )

        compiled_forward = torch.compile(
            attention_block.forward_varlen, backend="eager"
        )
        with OpLoggingMode() as mode:
            output = compiled_forward(x_packed, cu_seq, shape.max_seq_len)

            varlen_grad_out = torch.ones_like(output)
            _ = torch.autograd.grad(
                outputs=output,
                inputs=x_packed,
                grad_outputs=varlen_grad_out,
                retain_graph=True,
                create_graph=False,
                allow_unused=False,
            )[0]

        called_ops = mode.called_ops

        custom_ops_called = any(
            "torch_attn._varlen_attn" in op for op in called_ops
        ) and any("torch_attn._varlen_attn_backward" in op for op in called_ops)
        if not custom_ops_called:
            raise AssertionError("custom varlen attention ops should have been called")

        # Also verify _varlen_attn_out dispatches correctly under compile
        q, k, v = attention_block.get_varlen_qkv(x_packed.detach())

        def run_varlen_out(q, k, v, cu_seq, max_len):
            out_buf = torch.empty_like(q)
            varlen_attn_out(out_buf, q, k, v, cu_seq, cu_seq, max_len, max_len)
            return out_buf

        compiled_out = torch.compile(run_varlen_out, backend="eager")
        with OpLoggingMode() as out_mode:
            compiled_out(q, k, v, cu_seq, shape.max_seq_len)

        if not any("torch_attn._varlen_attn_out" in op for op in out_mode.called_ops):
            raise AssertionError("custom _varlen_attn_out op should have been called")

    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @parametrize(
        "sdpa_backend",
        ["aotriton", "ck"] if PLATFORM_SUPPORTS_CK_SDPA else ["aotriton"],
    )
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    @parametrize("scale", [None, 0.1])
    @parametrize(
        "window_size",
        [
            (-1, -1),
            (-1, 0),
            (4, 0),
            (-1, 4),
            (0, 0),
            (1, 0),
            (0, 1),
            (0, -1),
            (4, 4),
            (8, 2),
            (1025, 1025),
        ],
    )
    @parametrize(
        "backend",
        _varlen_backends(include_fa4_paged_kv=False),
    )
    @parametrize("enable_gqa", [False, True])
    def test_varlen_vs_sdpa(
        self, device, dtype, scale, window_size, backend, enable_gqa, sdpa_backend=None
    ):
        if TEST_WITH_ROCM:
            torch.backends.cuda.preferred_rocm_fa_library(sdpa_backend)

        torch.manual_seed(42)

        num_heads = 16
        num_kv_heads = 4 if enable_gqa else num_heads
        shape = VarlenShape(
            batch_size=4,
            max_seq_len=1024,
            embed_dim=1024,
            num_heads=num_heads,
            num_kv_heads=num_kv_heads,
        )

        batch_data = create_variable_length_batch(shape, device, dtype)
        seq_lengths = batch_data["seq_lengths"]
        cu_seq = batch_data["cu_seq"]
        max_len = batch_data["max_len"]
        x_packed = batch_data["x_packed"]
        x_padded = batch_data["x_padded"]
        x_padded_ref = batch_data["x_padded_ref"]

        golden_attention_block = AttentionBlock(
            shape.embed_dim,
            shape.num_heads,
            device,
            torch.float32,
            num_kv_heads=num_kv_heads,
        )
        attention_block = AttentionBlock(
            shape.embed_dim,
            shape.num_heads,
            device,
            dtype,
            num_kv_heads=num_kv_heads,
        )
        with torch.no_grad():
            attention_block.q_proj.weight.copy_(
                golden_attention_block.q_proj.weight.to(dtype)
            )
            attention_block.kv_proj.weight.copy_(
                golden_attention_block.kv_proj.weight.to(dtype)
            )
            attention_block.out_proj.weight.copy_(
                golden_attention_block.out_proj.weight.to(dtype)
            )

        with _use_backend(backend):
            varlen_output = attention_block.forward_varlen(
                x_packed,
                cu_seq,
                max_len,
                scale=scale,
                window_size=window_size,
            )
        sdpa_output = attention_block.forward_sdpa(
            x_padded,
            seq_lengths,
            scale=scale,
            window_size=window_size,
        )
        golden_sdpa_output = golden_attention_block.forward_sdpa(
            x_padded_ref,
            seq_lengths,
            scale=scale,
            window_size=window_size,
        )

        start_idx = 0
        for i, seq_len in enumerate(seq_lengths):
            end_idx = start_idx + seq_len

            varlen_seq = varlen_output[start_idx:end_idx]
            sdpa_seq = sdpa_output[i, :seq_len]
            golden_seq = golden_sdpa_output[i, :seq_len]

            fwd_atol = 2 * (golden_seq + 0.3 - 0.3 - golden_seq).abs().max().item()

            varlen_error = (varlen_seq - golden_seq.to(dtype)).abs().max().item()
            sdpa_error = (sdpa_seq - golden_seq.to(dtype)).abs().max().item()

            self.assertLessEqual(
                varlen_error,
                2 * sdpa_error + fwd_atol,
            )

            start_idx = end_idx

        with _use_backend(backend):
            grad_out = torch.randn_like(varlen_output)
            sdpa_grad_out = torch.zeros_like(sdpa_output)
            golden_sdpa_grad_out = torch.zeros(
                shape.batch_size,
                max_len,
                shape.embed_dim,
                device=device,
                dtype=torch.float32,
            )
            start_idx = 0
            for i, seq_len in enumerate(seq_lengths):
                end_idx = start_idx + seq_len
                sdpa_grad_out[i, :seq_len] = grad_out[start_idx:end_idx]
                golden_sdpa_grad_out[i, :seq_len] = grad_out[start_idx:end_idx].to(
                    torch.float32
                )
                start_idx = end_idx

            varlen_grad = torch.autograd.grad(
                outputs=varlen_output,
                inputs=x_packed,
                grad_outputs=grad_out,
            )[0]

        sdpa_grad = torch.autograd.grad(
            outputs=sdpa_output,
            inputs=x_padded,
            grad_outputs=sdpa_grad_out,
        )[0]

        golden_sdpa_grad = torch.autograd.grad(
            outputs=golden_sdpa_output,
            inputs=x_padded_ref,
            grad_outputs=golden_sdpa_grad_out,
        )[0]

        start_idx = 0
        for i, seq_len in enumerate(seq_lengths):
            end_idx = start_idx + seq_len

            varlen_grad_seq = varlen_grad[start_idx:end_idx]
            sdpa_grad_seq = sdpa_grad[i, :seq_len]
            golden_grad_seq = golden_sdpa_grad[i, :seq_len]

            bwd_atol = (
                2 * (golden_grad_seq + 0.3 - 0.3 - golden_grad_seq).abs().max().item()
            )

            varlen_error = (
                (varlen_grad_seq - golden_grad_seq.to(dtype)).abs().max().item()
            )
            sdpa_error = (sdpa_grad_seq - golden_grad_seq.to(dtype)).abs().max().item()

            self.assertLessEqual(
                varlen_error,
                2 * sdpa_error + bwd_atol,
            )

            start_idx = end_idx

    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    @parametrize("num_splits", [1, None])
    @parametrize(
        "window_size",
        [
            (-1, -1),
            (-1, 0),
            (1025, 1025),
            (384, 0),  # edge case
        ],
    )
    @parametrize(
        "backend",
        ["fa2"] + (["fa3"] if IS_SM90 else []) + (["fa4"] if SM100OrLater else []),
    )
    def test_batch_invariance(
        self, device, dtype, num_splits, window_size, backend, sdpa_backend=None
    ):
        if TEST_WITH_ROCM:
            torch.backends.cuda.preferred_rocm_fa_library(sdpa_backend)

        split_kwargs = {"num_splits": num_splits} if backend != "fa2" else {}

        torch.manual_seed(42)

        num_heads, head_dim = 2, 128
        target_seq_len = 512
        extra_seq_len = 1024

        target_q = torch.randn(
            target_seq_len, num_heads, head_dim, device=device, dtype=dtype
        )
        target_k = torch.randn(
            target_seq_len, num_heads, head_dim, device=device, dtype=dtype
        )
        target_v = torch.randn(
            target_seq_len, num_heads, head_dim, device=device, dtype=dtype
        )

        extra_q = torch.randn(
            extra_seq_len, num_heads, head_dim, device=device, dtype=dtype
        )
        extra_k = torch.randn(
            extra_seq_len, num_heads, head_dim, device=device, dtype=dtype
        )
        extra_v = torch.randn(
            extra_seq_len, num_heads, head_dim, device=device, dtype=dtype
        )

        cu_seq_solo = torch.tensor(
            [0, target_seq_len], device=device, dtype=torch.int32
        )
        cu_seq_batch = torch.tensor(
            [0, target_seq_len, target_seq_len + extra_seq_len],
            device=device,
            dtype=torch.int32,
        )

        all_q = torch.cat([target_q, extra_q], dim=0)
        all_k = torch.cat([target_k, extra_k], dim=0)
        all_v = torch.cat([target_v, extra_v], dim=0)

        # fa4 is batch invariant (num_splits=1) by default
        with _use_backend(backend), torch.no_grad():
            solo_output = varlen_attn(
                target_q,
                target_k,
                target_v,
                cu_seq_solo,
                cu_seq_solo,
                target_seq_len,
                target_seq_len,
                window_size=window_size,
                **split_kwargs,
            )

            batched_output = varlen_attn(
                all_q,
                all_k,
                all_v,
                cu_seq_batch,
                cu_seq_batch,
                extra_seq_len,
                extra_seq_len,
                window_size=window_size,
                **split_kwargs,
            )

            solo_out_buf = torch.empty_like(target_q)
            varlen_attn_out(
                solo_out_buf,
                target_q,
                target_k,
                target_v,
                cu_seq_solo,
                cu_seq_solo,
                target_seq_len,
                target_seq_len,
                window_size=window_size,
                **split_kwargs,
            )

            batched_out_buf = torch.empty_like(all_q)
            varlen_attn_out(
                batched_out_buf,
                all_q,
                all_k,
                all_v,
                cu_seq_batch,
                cu_seq_batch,
                extra_seq_len,
                extra_seq_len,
                window_size=window_size,
                **split_kwargs,
            )
            if num_splits == 1:
                self.assertEqual(solo_output, batched_output[:target_seq_len])
                self.assertEqual(solo_out_buf, batched_out_buf[:target_seq_len])
                self.assertEqual(solo_output, solo_out_buf)
            else:
                if backend == "fa3":
                    self.assertNotEqual(solo_output, batched_output[:target_seq_len])
                    self.assertNotEqual(solo_out_buf, batched_out_buf[:target_seq_len])

    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @unittest.skipIf(TEST_WITH_ROCM, "ROCm does not support seqused_k")
    @decorateIf(
        unittest.expectedFailure,
        lambda params: params["backend"] != "fa2"
        and any(kv_len < 128 for kv_len in params["actual_kv_lens"]),
    )
    @parametrize(
        "sdpa_backend",
        ["aotriton", "ck"] if PLATFORM_SUPPORTS_CK_SDPA else ["aotriton"],
    )
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    @parametrize(
        "actual_kv_lens",
        [
            [32, 64, 96, 48],
            [1, 1, 1, 1],
            [128, 128, 128, 128],
            [1, 128, 1, 128],
            [127, 63, 33, 17],
        ],
    )
    @parametrize("backend", _varlen_backends(include_fa4_paged_kv=False))
    def test_seqused_k_kv_cache(
        self, device, dtype, actual_kv_lens, backend, sdpa_backend=None
    ):
        if TEST_WITH_ROCM:
            torch.backends.cuda.preferred_rocm_fa_library(sdpa_backend)

        torch.manual_seed(42)

        batch_size = 4
        num_heads = 8
        head_dim = 64
        cache_size = 128

        q_seqs = [
            torch.randn(1, num_heads, head_dim, device=device, dtype=dtype)
            for _ in range(batch_size)
        ]
        q_packed, cu_seq_q, max_q = pack_sequences(q_seqs, device)

        k_seqs = [
            torch.randn(kv_len, num_heads, head_dim, device=device, dtype=dtype)
            for kv_len in actual_kv_lens
        ]
        v_seqs = [
            torch.randn(kv_len, num_heads, head_dim, device=device, dtype=dtype)
            for kv_len in actual_kv_lens
        ]

        k_cache_slots = []
        v_cache_slots = []
        for i in range(batch_size):
            k_slot = torch.full(
                (cache_size, num_heads, head_dim),
                float("nan"),
                device=device,
                dtype=dtype,
            )
            v_slot = torch.full(
                (cache_size, num_heads, head_dim),
                float("nan"),
                device=device,
                dtype=dtype,
            )
            k_slot[: actual_kv_lens[i]] = k_seqs[i]
            v_slot[: actual_kv_lens[i]] = v_seqs[i]
            k_cache_slots.append(k_slot)
            v_cache_slots.append(v_slot)

        k_cache_packed = torch.cat(k_cache_slots, dim=0)
        v_cache_packed = torch.cat(v_cache_slots, dim=0)
        cu_seq_k_cache = torch.arange(
            0,
            (batch_size + 1) * cache_size,
            cache_size,
            device=device,
            dtype=torch.int32,
        )
        seqused_k = torch.tensor(actual_kv_lens, device=device, dtype=torch.int32)

        with _use_backend(backend), torch.no_grad():
            output_cached = varlen_attn(
                q_packed,
                k_cache_packed,
                v_cache_packed,
                cu_seq_q,
                cu_seq_k_cache,
                max_q,
                cache_size,
                seqused_k=seqused_k,
            )

        k_real_packed, cu_seq_k_real, max_k_real = pack_sequences(k_seqs, device)
        v_real_packed = torch.cat(v_seqs, dim=0)

        with _use_backend(backend), torch.no_grad():
            output_reference = varlen_attn(
                q_packed,
                k_real_packed,
                v_real_packed,
                cu_seq_q,
                cu_seq_k_real,
                max_q,
                max_k_real,
            )

        self.assertFalse(output_cached.isnan().any())
        self.assertEqual(output_cached, output_reference)

        # varlen_attn_out with seqused_k should match
        with _use_backend(backend), torch.no_grad():
            out_buf = torch.empty_like(q_packed)
            output_out = varlen_attn_out(
                out_buf,
                q_packed,
                k_cache_packed,
                v_cache_packed,
                cu_seq_q,
                cu_seq_k_cache,
                max_q,
                cache_size,
                seqused_k=seqused_k,
            )
            self.assertEqual(output_out.data_ptr(), out_buf.data_ptr())
            self.assertEqual(out_buf, output_cached)

    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @unittest.skipIf(TEST_WITH_ROCM, "ROCm does not support seqused_k")
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    @parametrize("page_size", [32, 64, 128, 256])
    @parametrize("compile", [False, True])
    @parametrize(
        "actual_kv_lens",
        [
            [32, 64, 96, 48],
            [1, 1, 1, 1],
            [128, 128, 128, 128],
            [1, 128, 1, 128],
            [127, 63, 33, 17],
        ],
    )
    @parametrize(
        "backend",
        _varlen_backends(include_fa4_paged_kv=True),
    )
    def test_block_table_kv_cache(
        self, device, dtype, page_size, compile, actual_kv_lens, backend
    ):
        if backend == "fa2" and page_size % 256 != 0:
            self.skipTest("FA2 paged KV requires page_size divisible by 256")

        torch.manual_seed(42)

        batch_size = 4
        num_heads = 8
        head_dim = 64
        max_kv = max(actual_kv_lens)
        max_pages_per_seq = (max_kv + page_size - 1) // page_size
        cache_size = max_pages_per_seq * page_size
        total_pages = batch_size * max_pages_per_seq

        q_seqs = [
            torch.randn(1, num_heads, head_dim, device=device, dtype=dtype)
            for _ in range(batch_size)
        ]
        q_packed, cu_seq_q, max_q = pack_sequences(q_seqs, device)

        k_pages = torch.randn(
            total_pages, page_size, num_heads, head_dim, device=device, dtype=dtype
        )
        v_pages = torch.randn(
            total_pages, page_size, num_heads, head_dim, device=device, dtype=dtype
        )
        block_table = torch.randperm(
            total_pages, device=device, dtype=torch.int32
        ).view(batch_size, max_pages_per_seq)
        seqused_k = torch.tensor(actual_kv_lens, device=device, dtype=torch.int32)

        idx = (
            block_table.long()
            .view(-1, 1, 1, 1)
            .expand(-1, page_size, num_heads, head_dim)
        )
        k_gathered = k_pages.gather(0, idx).view(
            batch_size, cache_size, num_heads, head_dim
        )
        v_gathered = v_pages.gather(0, idx).view(
            batch_size, cache_size, num_heads, head_dim
        )
        k_seqs = [k_gathered[i, : actual_kv_lens[i]] for i in range(batch_size)]
        v_seqs = [v_gathered[i, : actual_kv_lens[i]] for i in range(batch_size)]

        k_real_packed, cu_seq_k_real, max_k_real = pack_sequences(k_seqs, device)
        v_real_packed = torch.cat(v_seqs, dim=0)

        attn_fn = torch.compile(varlen_attn, fullgraph=True) if compile else varlen_attn

        # Reference: no block_table
        with _use_backend(backend), torch.no_grad():
            output_reference = varlen_attn(
                q_packed,
                k_real_packed,
                v_real_packed,
                cu_seq_q,
                cu_seq_k_real,
                max_q,
                max_k_real,
            )

        cu_seq_k = torch.arange(
            0,
            (batch_size + 1) * cache_size,
            cache_size,
            device=device,
            dtype=torch.int32,
        )

        # FA2 requires cu_seq_k for paged KV; FA3/FA4 pass None
        cu_seq_k_paged = cu_seq_k if backend == "fa2" else None

        with _use_backend(backend), torch.no_grad():
            output_paged = attn_fn(
                q_packed,
                k_pages,
                v_pages,
                cu_seq_q,
                cu_seq_k_paged,
                max_q,
                cache_size,
                seqused_k=seqused_k,
                block_table=block_table,
            )

        self.assertEqual(output_paged, output_reference)

        # varlen_attn_out with paged KV cache should match
        with _use_backend(backend), torch.no_grad():
            out_buf = torch.empty_like(q_packed)
            output_out = varlen_attn_out(
                out_buf,
                q_packed,
                k_pages,
                v_pages,
                cu_seq_q,
                cu_seq_k_paged,
                max_q,
                cache_size,
                seqused_k=seqused_k,
                block_table=block_table,
            )
            self.assertEqual(output_out.data_ptr(), out_buf.data_ptr())
            self.assertEqual(out_buf, output_paged)

        # compile the lower level aten op (FA3 only, will cause graph break)
        if compile and backend != "fa2":
            compiled_aten_op = torch.compile(
                torch.ops.aten._flash_attention_forward_no_dropout_inplace
            )
            with _use_backend(backend), torch.no_grad():
                out_buf = torch.empty_like(q_packed)
                compiled_aten_op(
                    out_buf,
                    q_packed,
                    k_pages,
                    v_pages,
                    cu_seq_q,
                    None,
                    max_q,
                    cache_size,
                    0.0,
                    False,
                    False,
                    seqused_k=seqused_k,
                    block_table=block_table,
                )
            self.assertEqual(out_buf, output_reference)

    @unittest.skipIf(
        not PLATFORM_SUPPORTS_FLASH_ATTENTION, "Flash Attention not supported"
    )
    @parametrize("dtype", [torch.bfloat16, torch.float16])
    @parametrize(
        "backend",
        ["fa2"] + (["fa3"] if IS_SM90 else []) + (["fa4"] if SM100OrLater else []),
    )
    def test_enable_gqa(self, device, dtype, backend):
        torch.manual_seed(42)

        head_dim = 64
        seq_len = 512
        num_heads_q, num_heads_k = 16, 4
        total_tokens = 2 * seq_len

        q = torch.randn(total_tokens, num_heads_q, head_dim, device=device, dtype=dtype)
        k = torch.randn(total_tokens, num_heads_k, head_dim, device=device, dtype=dtype)
        v = torch.randn(total_tokens, num_heads_k, head_dim, device=device, dtype=dtype)
        cu_seq = torch.tensor(
            [0, seq_len, total_tokens], device=device, dtype=torch.int32
        )

        with self.assertRaisesRegex(ValueError, "enable_gqa=True"):
            varlen_attn(q, k, v, cu_seq, cu_seq, seq_len, seq_len)

        with self.assertRaisesRegex(ValueError, "enable_gqa=True"):
            varlen_attn_out(
                torch.empty_like(q), q, k, v, cu_seq, cu_seq, seq_len, seq_len
            )

        k_bad = torch.randn(total_tokens, 3, head_dim, device=device, dtype=dtype)
        v_bad = torch.randn(total_tokens, 3, head_dim, device=device, dtype=dtype)
        with self.assertRaisesRegex(ValueError, "multiple of kv heads"):
            varlen_attn(
                q, k_bad, v_bad, cu_seq, cu_seq, seq_len, seq_len, enable_gqa=True
            )

        with _use_backend(backend), torch.no_grad():
            out = varlen_attn(
                q, k, v, cu_seq, cu_seq, seq_len, seq_len, enable_gqa=True
            )
            out_buf = torch.empty_like(q)
            varlen_attn_out(
                out_buf, q, k, v, cu_seq, cu_seq, seq_len, seq_len, enable_gqa=True
            )
            self.assertEqual(out_buf, out)


device_types = ("cuda",)

instantiate_device_type_tests(TestVarlenAttention, globals(), only_for=device_types)

if __name__ == "__main__":
    run_tests()
